# $Id: functions.R,v 1.3 2011/09/14 14:09:36 srevilak Exp $

my.cex <- 1.5

data.dir <- "~/Documents/schoolwork/phd/si/mysql-test-results/par-cooked"

## Prefer fixed to scientific notation.
options(scipen=5)

## # write the current graphics device to a file
## #
## # This routine uses the file extension to select the
## # output routine.  
## write.device <- function(outfile) {
## 
##   if (regexpr("\\.png$", outfile) >= 0) {
##     dev.copy(png, file=outfile, width=800, height=600);
##   }
## 
##   else if (regexpr("\\.pdf$", outfile) >= 0) {
##     dev.copy(pdf, file=outfile, # width=8, height=6,
##              pointsize=12,
##              paper="USr", width=10, height=7.5,
##              useDingbats=FALSE)
##   }
##   else {
##     stop("image extension not handled in ", outfile)
##   }
##   dev.off();
## }


##################################################################
# Types
##################################################################

## S4 object to hold information about a data set ("dset")
## 
setClass("dset",
         representation(id="character",
                        sequence="integer",
                        file="character",
                        data="data.frame",
                        notes="character",
                        skun="character",
                        isolation="character",
                        hotspot="integer"));


####################################################################
## Standard IO Stuff
####################################################################

my.read <- function(filename, is.numeric=TRUE) {
  if (is.numeric) {
    d <- read.table(filename, header=TRUE, sep="\t", colClasses="numeric");
  }
  else {
    d <- read.table(filename, header=TRUE, sep="\t");
  }
  d;
}


##
## Given data set "d", determine skun
##
determine.skun <- function(d) {
  reads <- d[['config.select.statements']][1]
  writes <- d[['config.update.statements']][1]
  if (reads > 0 && writes > 0) {
    return (sprintf("s%du%d", reads, writes));
  }
  stop("find.skun: did not find reads, writes:", reads, writes);
}


##
## Given data set "d", determine hotspot size
##
determine.hotspot <- function(d) {
  pct.hot <- d[['config.hotspot.size']][1]
  if (pct.hot > 0) {
    return (as.integer(pct.hot * 1000000));
  }
  stop("determine.hotspot: did not find config.hotspot.size: ", pct.hot);
}


##
## given data set "d", determine isolation leve
##
determine.isolation <- function(d) {
  comment <- d[['version_comment']][1];
  dstest <- d[['innodb_dangerous_structure_testing']][1]
  iso <- d[['tx_isolation']][1]

  # comment in ("umb-non-debug-gcc4.5", "dist-mysql")
  
  if (regexpr("dist-mysql", comment) == 1 &&
      iso == "SERIALIZABLE") {
    return ("S2PL");
  }

  # our first few runs had the wrong version_comment in the umb build
  if (regexpr("umb-non-debug", comment) < 0 &&
      regexpr("umb-debug", comment) < 0) {
    stop("comment not matched: ", paste(comment, dstest, iso));
  }

  if (iso == "REPEATABLE-READ") {
    return ("SI");

  }

  else if (iso == "SERIALIZABLE" && dstest == "OFF") {
    return ("PSSI");

  }

  else if (iso == "SERIALIZABLE" && dstest == "ON") {
    return ("ESSI");
  }

  stop ("cannot determine isolation level:", comment, iso, dstest);
}



##
## reads a set of files and return a list of
## "dset" objects.
##
## filedata is a list of pairs.
##
## Returns a list of dset objects
##
read.files <- function(dir=".", filenames) {

  outfiles <- list();
  counter = 1;

  for (f in filenames) {
    seq <- as.integer(gsub("\\.[^.]*$", "", f))
    d <- my.read(file.path(dir, f), is.numeric=FALSE);

    rownames(d) <- d[['MPL']]

    dset <- new("dset")
    dset@data <- d;
    dset@file <- f;
    dset@id <- f;
    dset@sequence <- seq
    dset@skun <- determine.skun(d);
    dset@hotspot <- determine.hotspot(d);
    dset@isolation <- determine.isolation(d);

    outfiles[counter] <- dset
    counter <- counter + 1
  }

  outfiles
}


##
## given list of dsets, make one big table
##
to.table <- function(dlist) {
  df <- data.frame()

  for (dset in dlist) {
    d <- dset@data
    colcount <- ncol(d)
    rowcount <- nrow(d)
    sequence <- rep(dset@sequence, rowcount);
    isolation.level <- rep(dset@isolation, rowcount);
    skun <- rep(dset@skun, rowcount);
    hotspot <- rep(dset@hotspot, rowcount);
    skun.h <- sprintf("%s-%d", skun, hotspot);
    mpl <- d['MPL'];
    not.mpl <- !(names(d) %in% "MPL")
    tmp.df <- d[not.mpl]
    tmp.df <- cbind(mpl, sequence, isolation.level, skun, hotspot, skun.h, tmp.df)

    df <- rbind(df, tmp.df)
  }

  total.rows <- nrow(df);
  rownames(df) <- 1:total.rows
  df
}


##
## write a table to a data file 
##
write.data <- function(file, data) {

  fd <- file(description=file, open="w+")

  write.table(data, file=fd, quote=FALSE, sep="\t", row.names=FALSE, na="")

  close(fd)  
}



####################################################################
## Computation of Derived Data
####################################################################

##
## given "results" (a list of dsets), add a standard set of
## derived measures
##
add.derived.measures <- function(results) {
  results <- add.ctps(results)
  results <- add.pct.failed(results)
  results <- add.avg.commited.tx.duration(results)
  results <- add.avg.commit.duration(results)
  results <- add.avg.exec.duration(results)
  results <- add.commits.per.flush(results);
  results <- add.pct.fuw(results);
  results <- add.pct.sr.err(results)
  results <- add.avg.ctg.size(results)
  results <- add.avg.prune.nodes.examined(results);
  results <- add.cycle.edges.traversed(results);
  results <- add.avg.cycle.length(results);
  
  results;
}


add.ctps <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "ctps=tx.committed/avg.thread.duration.sec");
    df <- d@data
    ctps <- df['tx.committed']/df['avg.thread.duration.sec']
    d@data['ctps'] <- ctps
    d
  })
}

add.fuw.sec <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "fuw_sec=errors.fcw/avg.thread.duration.sec");
    df <- d@data
    fuw <- df['errors.fcw']/df['avg.thread.duration.sec']
    d@data['fuw_sec'] <- fuw
    d
  })
}


add.sr.err.sec <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "sr_error_sec=errors.cycle/avg.thread.duration.sec");
    df <- d@data
    sr.error <- df['errors.cycle']/df['avg.thread.duration.sec']
    d@data['sr_error_sec'] <- sr.error
    d
  })
}


add.log.flush.rate <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "log_flush_sec=Innodb_os_log_fsyncs/(adjusted avg.thread.duration.sec)");
    df <- d@data

    warmup <- 70
    cooldown <- 5
    if ("warmupPeriod" %in% names(df)) {
      warmup <- df['warmupPeriod']
      cooldown <- warmup;
      if (cooldown > 5) {
        cooldown <- 5;
      }
    }
    syncs <- df['Innodb_os_log_fsyncs'] /
      (df['avg.thread.duration.sec'] + warmup + cooldown)
    d@data['log_flush_sec'] <- syncs
    d;
  })
}


add.commits.per.flush <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d,
                   paste("commits_per_flush=",
                         "(tx-committed)/",
                         "(adjusted Innodb_os_log_fsyncs)",
                         sep=""));
    df <- d@data

    warmup <- 70
    cooldown <- 5
    if ("warmupPeriod" %in% names(df)) {
      warmup <- df['warmupPeriod']
      cooldown <- warmup;
      if (cooldown > 5) {
        cooldown <- 5;
      }
    }

    scale =  (df['avg.thread.duration.sec'] /
              (df['avg.thread.duration.sec'] + warmup + cooldown));

    commits <- df['tx.committed'];
    adj.flush <- df['Innodb_os_log_fsyncs'] * scale;
                  

    d@data['commits_per_flush'] <- commits / adj.flush
    d;
  })
}


add.pct.failed <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "pct_failed=100*(tx.rollback/total.transactions)");
    df <- d@data
    pct.fail <- 100.0 * (df['tx.rollback']/df['total.transactions'])
    d@data['pct_failed'] <- pct.fail
    d
  })
}

add.pct.fuw <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "pct_fuw=100*(errors.fcw/total.transactions)");
    df <- d@data
    pct.fuw <- 100.0 * (df['errors.fcw']/df['total.transactions'])
    d@data['pct_fuw'] <- pct.fuw
    d
  })
}

add.pct.sr.err <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "pct_sr_err=100*(errors.cycle/total.transactions)");
    df <- d@data
    pct.sr.err <- 100.0 * (df['errors.cycle']/df['total.transactions'])
    d@data['pct_sr_err'] <- pct.sr.err
    d
  })
}

add.avg.commit.duration <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "avg.commit.duration=sum.commit.duration.ms/tx.committed");
    df <- d@data;
    sum.duration <- df['sum.commit.duration.ms']
    tx.committed <- df['tx.committed']
    d@data['avg.commit.duration'] <- (sum.duration / tx.committed)
    d
  })
}


# Average duration of committed transactions, including the commit
add.avg.commited.tx.duration <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, paste("avg.commit.duration=",
                            "(sum.commit.duration.ms+sum.exec.duration.ms)",
                            "/tx.committed", sep=""));
    df <- d@data;
    total.duration <- df['sum.commit.duration.ms'] + df['sum.exec.duration.ms']
    tx.committed <- df['tx.committed']
    d@data['avg.commited.tx.duration'] <- (total.duration / tx.committed)
    d
  })
}

add.avg.exec.duration <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, paste("avg.exec.duration=",
                            "sum.exec.duration.ms",
                            "/tx.committed", sep=""));
    df <- d@data;
    exec.duration <- df['sum.exec.duration.ms'] / df['tx.committed']
    d@data['avg.exec.duration'] <- exec.duration
    d
  })
}

add.pct.time.in.commit <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, paste("pct.time.in.commit=",
                            "100 * (sum.commit.duration.ms)/",
                            "(sum.commit.duration.ms + sum.exec.duration.ms)",
                            sep=""));
    df <- d@data;
    total.duration <- df['sum.commit.duration.ms'] + df['sum.exec.duration.ms']
    commit.duration <- df['sum.commit.duration.ms']
    d@data['pct.time.in.commit'] <- 100 * (commit.duration / total.duration)
    d
  })
}


add.avg.ctg.size <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, paste("avg.ctg.size=",
                            "(Innodb_sum_ctg_size/Innodb_cycle_tests_done)",
                            sep=""));
    df <- d@data;
    d@data['avg.ctg.size'] <- df['Innodb_sum_ctg_size'] /
      df['Innodb_cycle_tests_done'] 
    d
  })
}


add.avg.prune.nodes.examined <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, paste("avg.prune.nodes.examined=",
                            "(Innodb_prune_nodes_examined/Innodb_prune_count)",
                            sep=""));
    df <- d@data;
    d@data['avg.prune.nodes.examined'] <- df['Innodb_prune_nodes_examined'] /
      df['Innodb_prune_count'] 
    d
  })
}


# number of edges traversed during cycle testing.
add.cycle.edges.traversed <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, paste("avg.cycle.edges.traversed=",
                            "(Innodb_ctg_edges_traversed/Innodb_cycle_tests_done)",
                            sep=""));
    df <- d@data;
    d@data['avg.cycle.edges.traversed'] <- df['Innodb_ctg_edges_traversed'] /
      df['Innodb_cycle_tests_done'] 
    d
  })
}


add.avg.cycle.length <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "avg.cycle.length=Innodb_sum_cycle_length/Innodb_cycles_found");
    df <- d@data
    avg.len <- df['Innodb_sum_cycle_length']/df['Innodb_cycles_found']
    d@data['avg.cycle.length'] <- avg.len
    d
  })
}




# Adds note to a dset.  This is typically used to keep track
# of provenance.
#
dset.note <- function(d, note) {
  idx <- 1+ length(slot(d, 'notes'));
  slot(d, 'notes')[idx] <- note
  d
}



####################################################################
## Data Summarization
####################################################################


collect.measure <- function(data, measure) {
  first <- data[[1]]
  if (class(first)[1] != "dset") {
    stop("data is not a list of dset objects")
  }
  df <- first@data;
  dnames <- names(df)

  if (! (measure %in% dnames)) {
    namelist <- paste(dnames, sep=",")
    stop("no ", measure, " in data frame")
  }
  
  outf <- data.frame(MPL=df['MPL'])
  sequences <- numeric()
  counter <- 1
  
  for (d in data) {
    id <- d@id
    datum <- d@data[measure]
    outf[id] <- datum
    sequences[counter] <- d@sequence
    counter <- counter + 1
  }
  attr(outf, "sequences") <- sequences
  attr(outf, "measure") <- measure
  
  outf
}


